{{
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DS1307 Real Time Clock Engine
//
// This is a dervivitave of the original, using c3io to replace
// the low level sections.
//
// Adapted to c3io by: Joe Hohertz
// Author: Kwabena W. Agyeman
// Updated: 7/27/2010
// Designed For: P8X32A
// Version: 1.4
//
// Copyright (c) 2010 Kwabena W. Agyeman
// See end of file for terms of use.
//
// Update History:
//
// v1.0 - Original release - 2/8/2009.
// v1.1 - Made code faster - 4/18/2009.
// v1.2 - Included error checking code - 2/27/2010.
// v1.3 - Added variable pin setup and locks - 4/17/2010.
// v1.4 - Fixed possible user error bugs - 7/27/2010.
//
// For each included copy of this object only one spin interpreter should access it at a time.
//
// Nyamekye,
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// I2C Circuit:
//
//                  3.3V
//                   |
//                   R 10KOHM
//                   |
// Data Pin Number  --- DS1307 SDA Pin.
//
//                  3.3V
//                   |
//                   R 10KOHM
//                   |
// Clock Pin Number --- DS1307 SCL Pin.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}

CON

  #1, Sunday, Monday, Tuesday, Wednesday, Thursday, Friday, Saturday
  #1, January, February, March, April, May, June, July, August, September, October, November, December

VAR

  byte timeTemporary, time[7]
  byte buffer[7]

OBJ

    io: "c3io"

PUB clockSecond '' 3 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the cached second (0 - 59) from the real time clock.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  return time

PUB clockMinute  '' 3 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the cached minute (0 - 59) from the real time clock.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  return time[1]

PUB clockHour '' 3 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the cached hour (0 - 23) from the real time clock.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  return time[2]

PUB clockDay '' 3 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the cached day (1 - 7) from the real time clock.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  return time[3]

PUB clockDate '' 3 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the cached date (1 - 31) from the real time clock.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  return time[4]

PUB clockMonth '' 3 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the cached month (1 - 12) from the real time clock.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  return time[5]

PUB clockYear '' 3 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the cached year (2000 - 2099) from the real time clock.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  return (time[6] + 2_000)

PUB clockMeridiemHour '' 6 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the cached meridiem hour (12 - 11) from the real time clock.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  result := (clockHour // 12)
  ifnot(result)
    result += 12

PUB clockMeridiemTime '' 6 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns true if the cached meridiem hour is post meridiem and false if the meridiem cached hour is ante meridiem.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  return (clockHour => 12)

PUB clockSquareWaveOut(frequency, state) '' 14 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Changes the current real time clock square wave output driver settings. Returns true on success and false on failure.
'' //
'' // Frequency - Selects the frequency of the square wave pin. (0 - 1HZ), (1 - 4.096KHZ), (2 - 8.192KHZ), (3 - 32.768KHZ).
'' // State - Selects the state of the square wave pin. (0 - SQW Off pin low), (1 - SQW On default low).
'' //                                                   (2 - SQW Off pin high), (3 - SQW On default high).
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  state := ((state <# 3) #> 0)
  return setRAM(7, (((state & $2) << 6) | (state & $1) | ((frequency <# 3) #> 0)))

PUB readTime | addr, index
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Caches the current real time clock settings. Returns true on success and false on failure.
'' //
'' // Call "checkSecond" to get the real time clock second after calling this function first.
'' // Call "checkMinute" to get the real time clock minute after calling this function first.
'' // Call "checkHour" to get the real time clock hour after calling this function first.
'' // Call "checkDay" to get the real time clock day after calling this function first.
'' // Call "checkDate" to get the real time clock date after calling this function first.
'' // Call "checkMonth" to get the real time clock month after calling this function first.
'' // Call "checkYear" to get the real time clock year after calling this function first.
'' // Call "checkMeridiemHour" to get the real time clock meridiem hour after calling this function first.
'' // Call "checkMeridiemTime" to get the real time clock meridiem time after calling this function first.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  addr := constant(%11010000 << 8)
  io.setLock
  io.doCommand((io#i2cRead1Cmd | io#ioLowSpeed), addr, @buffer, 7)
  io.clearLock

  bytefill(@time, 0, 7)
  repeat index from 0 to 6
    timeTemporary := buffer[index]
    time[index] := (((timeTemporary >> 4) * 10) + (timeTemporary & $F))
'    time[index] := timeTemporary
  return true

PUB writeTime(second, minute, hour, day, date, month, year) | index, addr
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Changes the current real time clock time settings. Returns true on success and false on failure.
'' //
'' // Second - Number to set the second to between 0 - 59.
'' // Minute - Number to set the minute to between 0 - 59.
'' // Hour - Number to set the hour to between 0 - 23.
'' // Day - Number to set the day to between 1 - 7.
'' // Date - Number to set the date to between 1 - 31.
'' // Month - Number to set the month to between 1 - 12.
'' // Year - Number to set the year to between 2000 - 2099.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  buffer := ((second <# 59) #> 0)
  buffer[1] := ((minute <# 59) #> 0)
  buffer[2] := ((hour <# 23) #> 0)
  buffer[3] := ((day <# 7) #> 1)
  buffer[4] := ((date <# 31) #> 1)
  buffer[5] := ((month <# 12) #> 1)
  buffer[6] := (((year <# 2_099) #> 2_000) - 2_000)
  ' make BCD
  repeat index from 0 to 6
    timeTemporary := buffer[index]
    buffer[index] := ((timeTemporary / 10) << 4) + (timeTemporary // 10)
  ' write it
  addr := constant(%11010000 << 8)
  io.setLock
  io.doCommand((io#i2cWrite1Cmd | io#ioLowSpeed), addr, @buffer, 7)
  io.clearLock
  return true

PUB readSRAM(index) '' 12 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Returns the selected SRAM value at the specified index on sucess and false on failure.
'' //
'' // Index - The byte location in SRAM to check. (0 - 55).
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  return getRam(((index <# 55) #> 0) + 8)

PUB writeSRAM(index, value) '' 14 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Changes the selected SRAM value at the specified index to the specified value.
'' //
'' // Returns true on success and false on failure.
'' //
'' // Index - The byte location in SRAM to change. (0 - 55).
'' // Value - The value to change the byte location to. (0 - 255).
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  return setRam((((index <# 55) #> 0) + 8), ((value <# 255) #> 0))

PUB pauseForSeconds(number) '' 4 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Pauses execution for a number of seconds.
'' //
'' // Number - Number of seconds to pause for between 0 and 2,147,483,647.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  result := cnt

  repeat (number #> 0)

    result += clkfreq
    waitcnt(result)

PUB pauseForMilliseconds(number) '' 4 Stack Longs
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Pauses execution for a number of milliseconds.
'' //
'' // Number - Number of milliseconds to pause for between 0 and 2,147,483,647.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  result := cnt
  repeat (number #> 0)
    result += (clkfreq / 1_000)
    waitcnt(result)

PRI setRAM(index, value) | addr ' 9 Stack Longs
'' if index > 0 then it does a rather standard I2C set of the value
'' if index = 0 then it treats value as a pointer to a 7 value array
'' which is will be encoded as BCD, and sent out
  addr := constant(%11010000 << 8) | index
  buffer[0] := value
  io.setLock
  io.doCommand((io#i2cWrite1Cmd | io#ioLowSpeed), addr, @buffer, 1)
  io.clearLock
  return true
'  result := (io.isError <> 0)                       ' Return any error code

PRI getRAM(index) | addr ' 8 Stack Longs
  addr := (%11010000 << 8) | index
  io.setLock
  io.doCommand((io#i2cRead1Cmd | io#ioLowSpeed), addr, @buffer, 1)
  io.clearLock
  return buffer[0]

DAT

' //////////////////////Variable Array/////////////////////////////////////////////////////////////////////////////////////////

lockNumber              byte 00 ' Driver lock number.

' /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

{{

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                  TERMS OF USE: MIT License
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
// Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}
