{
/////////////////////////////////////////////////////////////////////////////

FILENAME: c3po_drv_sd.spin

AUTHOR: Joe Hohertz
LAST MODIFIED: 2011-03-16
VERSION 0.1
LICENSE: MIT


This is an SD card "block" driver, which depends on c3io, both of which are
derived from 'jtc_c3_sdspiFemto.spin'. See c3io.spin with which this
file is always to be distributed for detailed license info.

TODO LIST:

- check mount state prior to trying other operations
- locking!
- efficiency improvements (not using multi-byte IO much/at all yet)


Original credits to:

{{
Modified Tiny Basic for use with Propeller Demo Board and Hydra.
I2C and SPI driver initialization & interface object derived from Propeller OS.

-Modified(hacked)SPIN routines in from Andre' Lamothe's C3SD_DEMO_010.SPIN
  and disabled ASM routines for minimal support for C3 boardn -JT Cook
Copyright (c) 2009 Michael Green.  See end of file for terms of use.
}}

}

CON

   ' SD card (secure digital) constants (you can reduce these to speed things up and see if the SD card still works)
   SD_RESPONSE_LIMIT        = 256     ' maximum number of iterations that SD card will be querried for response after a command sent
   SD_POWERUP_LIMIT         = 256     ' maximum number of iterations to initially clock SD card to power up
   SD_RETRIEVE_SECTOR_LIMIT = 4096    ' maximum number of iterations for SD to retrieve a requested sector/block
   SD_WRITE_SECTOR_LIMIT    = 128     ' maximum number of iterations for SD to write a requested sector/block
   SD_BUSY_LIMIT            = 128     ' maximum number of iterations for SD to wait in a busy loop

   ' SD commands (based on SD spec names)
   SD_GO_IDLE_STATE     = ($40 + 0)        ' resets the SD card, places it into SPI mode, idle
   MMC_SEND_OP_COND     = ($40 + 1)         ' takes SD out of idle state ready for commands
   SD_SEND_IF_COND      = ($40 + 8)   ' part of card initialization
   SD_SET_BLOCK_LENGTH  = ($40 + 16)  ' part of card init

   SD_READ_CSD          = ($40 + 9) ' Read the CSD register
   SD_READ_CID          = ($40 + 10) ' Read the CID register
   SD_READ_OCR          = ($40 + 58) ' Read the OCR register

   SD_APP_CMD           = ($40 + 55) ' makes sure card is in SD mode

   SD_SEND_STATUS       = ($40 + 16)  ' part of card init

   SD_SEND_OP_COND      = ($40 + 41) ' makes sure card is in SD mode
   SD_READ_SINGLE_BLOCK = ($40 + 17) ' reads a single block (usually a 512 byte sector) at byte address sent in command address
   SD_WRITE_SINGLE_BLOCK= ($40 + 24) ' write a single block (usually a 512 byte sector) at byte address sent in command address

   HCSBitMask        =    %01_000000_00000000_00000000_00000000              ' Parameter bit mask for HCS bit.
   OCRCheckMask      =    %00_000000_00110000_00000000_00000000              ' Parameter check mask for OCR bits.


   SECTOR_SIZE          = 512         ' should always be 512 for SD cards
   MAX_SECTORS_BUFFER   = 1           ' number of sectors buffer can hold, make larger if needed


VAR

long cardSize, readTimeout, writeTimeout
byte cardSpecificData[16] ' The secure digital card identification number.
byte response_token[5]      ' maximum 5 byte response
byte cardUniqueID[17] ' The secure digital card identification number.
byte SD_Address_Shift              ' an offset used to support SDHC vs. not addressing
byte cardMounted
byte isHC

OBJ

    io : "c3io"


PUB init
  ' tries for each type of operation
  readTimeout := (clkfreq / constant(8 * 8 * 10 * 100))
  writeTimeout := (clkfreq / constant(8 * 8 * 4 * 100))

PUB getCardSize
  return cardSize

pub getUniqueID(buffer) | index
  repeat index from 0 to 15
    byte[buffer][index] := cardUniqueID[index]

pub getCSD(buffer) | index
  repeat index from 0 to 15
    byte[buffer][index] := cardSpecificData[index]

PUB mount | temp, temp2, index, sd_response, cardType

  setLock

  io.SPI_Select_Channel( 0 )
  ' now clock the SD SPI interface a few times and let it initialize (send dummy data)
  repeat index from 0 to SD_POWERUP_LIMIT
    io.SPI_Send_Recv_Byte( $FF )

  ' assert chip select
  io.SPI_Select_Channel( 5 )

  ifnot Idle_Enter
      clearLock
      return ($FF)

  temp := Interface_Condition
  if Interface_Condition == 0
      clearLock
      return ($FF)

  cardType := 2          ' assume least common denominator (MMC)


  'Verify SD card mode (not a MMC)...
  ' send SD command APP_CMD = $40 + $37
  sd_response := SD_Send_Command( SD_APP_CMD, 0, $95, 1 )

  if ( sd_response <> 0 )
    ' Probably an MMC card
    ifnot Idle_Exit_MMC
        clearLock
        return ($FF)
  else
    ifnot Idle_Exit_SD
        clearLock
        return ($FF)
    cardType--          ' SD card, I guess? :)


  sd_response := SD_Send_Command( SD_READ_OCR, 0, $95, 5 )
  if sd_response <> 0
      clearLock
      return ($FF)
  temp := response_token[1] << 24
  temp |= response_token[2] << 16
  temp |= response_token[3] << 8
  temp |= response_token[4]

  if (temp & OCRCheckMask)       ' voltage check
    if (temp & $80000000)          ' ready check
      if (temp & $40000000)
        SD_Address_Shift := 0       ' HC has no shift.
        isHC := true
      else
        ' non-hc, shift address left by 9 bits on read/writes
        SD_Address_Shift := 9
        isHC := false

      ifnot isHC
        ' set the block length
        sd_response := SD_Send_Command( SD_SET_BLOCK_LENGTH, 512, $95, 1 )
        if sd_response <> 0
          clearLock
          return ($FF)

      ' read the CSD register
      sd_response := SD_Send_Command( SD_READ_CSD, 0, $95, 1 )
      if sd_response <> 0
        clearLock
        return ($FF)
      sd_response := SD_Read_Data_Block(@cardSpecificData, 16)
      if sd_response <> 0
        clearLock
        return ($FF)
      ' read the CID register
      sd_response := SD_Send_Command( SD_READ_CID, 0, $95, 1 )
      if sd_response <> 0
        return ($FF)
      sd_response := SD_Read_Data_Block(@cardUniqueId, 16)
      if sd_response <> 0
        clearLock
        return ($FF)

      cardSize   := 0                 ' Clear card size.

'      if ((cardType == 2) | ((cardSpecificData[0] & (1<<6)) == 0))
      ifnot isHC
        ' for MMC and low capacity SD...
        temp := (cardSpecificData[6] & $3) << 10
        temp |= cardSpecificData[7] << 2
        temp |= cardSpecificData[8] >> 6
        cardSize := temp
        temp := (cardSpecificData[9] & $3) << 1
        temp |= cardSpecificData[10] >> 7                ' cardSizeMultiplier
        temp2 := cardSpecificData[5] & $F                ' cardReadBlockLength
        temp2 -= 9
        temp  += 2
        cardSize += 1
        cardSize <<= temp2
        cardSize <<= temp
      else
        ' for high capacity
        temp := (cardSpecificData[7] & $3F) << 16
        temp |= cardSpecificData[8] << 8
        temp |= cardSpecificData[9]
        temp += 1
        temp <<= 10
        cardSize := temp

      if cardSize > $7FFFFFFF
        cardSize := $7FFFFFFF

      cardMounted := -1
      'SD card mounted ok
      clearLock
      return(0)' all ok

    ' end ready check
  ' end voltage check
  clearLock
  return ($FF)


PUB unmount
  ' FIXME: much more to cleanup here
  bytefill(@cardUniqueID, 0, 17)
  bytefill(@cardSpecificData, 0, 16)
  cardSize   := 0                 ' Clear card size.

' reading and writing sectors

PUB SD_Read_Sector(sector32, sectorbuffer16) | response

' DESCRIPTION: this function reads a sector from the SD card into the sent sector buffer
'
' INPUTS: sector32     - 32-bit absolute sector number to read
'         sectorbuffer - 16-bit pointer to receive buffer for sector
'
' OUTPUTS: storage pointed to by sectorbuffer16 will be filled with byte data from sector
'          returns $00 if no errors, $FF if error
' -----------------------------------------------------------------------------

  ' limit the sector choice to valid range
  ' FIXME: should probably just error out.
  if sector32 > (cardSize-1)
    sector32 := cardSize - 1

  ' HC cards like a byte address vs sector, adjust for it if needed
  sector32 <<= SD_Address_Shift

  setLock

  ' send the read block command with the byte address.
  response := SD_Send_Command( SD_READ_SINGLE_BLOCK , sector32, $95, 1 )
  ' test response, should be $00
  if (response <> $00)
    clearLock
    return (response)

  response := SD_Read_Data_Block(sectorbuffer16, SECTOR_SIZE)

  clearLock
  return (response)

' end SD_Read_Sector


PUB SD_Write_Sector(sector32, sectorbuffer16) | response
  ' limit the sector choice to valid range
  ' FIXME: should probably just error out.
  if sector32 > (cardSize-1)
    sector32 := cardSize - 1

  ' non-HC cards like a byte address vs sector, adjust for it if needed
  sector32 <<= SD_Address_Shift

  setLock

  ' send the read block command with the byte address.
  response := SD_Send_Command( SD_WRITE_SINGLE_BLOCK , sector32, $95, 1 )
  ' test response, should be $00
  if (response <> $00)
    clearLock
    return (response)

  response := SD_Write_Data_Block(sectorbuffer16, SECTOR_SIZE)
  if (response <> $00)
    clearLock
    return (response)

  ' wait around to see how it turned out...
  response := SD_Wait_Write_Busy
  if (response <> $00)    ' timed out
    clearLock
    return ($FF)

' FIXME: find out why this is failing?
'  ' check status register for issues
'  response := SD_Send_Command( SD_SEND_STATUS , sector32, $95, 2 )
'  if (response_token[0] <> 0) | (response_token[1] <> 0)
'    clearLock
'    return ($FF)

  ' all is hunky dory
  clearLock
  return (0)

' wrappers of above for fsrw
' FIXME: just tell fsrw to call the other...
PUB readSDCard(addr,buffer,count)                      '' Read block(s) from SD card to RAM
 SD_Read_Sector(addr, buffer)
PUB writeSDCard(addr,buffer,count)                     '' Write block(s) to SD card from RAM
 SD_Write_Sector(addr, buffer)


' local version of io locks that select and deselect the bus
PRI setLock
  io.setLock
  io.SPI_Select_Channel(io#SPI_CHAN_SD)
PRI clearLock
  io.SPI_Select_Channel(io#SPI_CHAN_NULL)
  io.SPI_Send_Recv_Byte( $FF )        ' sd needs extra clocks to clear the line
  io.clearLock


PRI SD_Send_Command( command8, address32, crc, response_length ) | response, index, index2

' DESCRIPTION: This function is the main workhorse of the SD interface and is used to send commands
' to the SD card via the SPI interface. All SD commands consist of 6 bytes in the following format:
'
' Command id     (1-byte) This you get from the SD specification.
' 32-bit address (4-bytes) In big endian form.
' CRC result     (1-byte) Error checking byte.
'
' The address is in big endian form, that is high byte to low byte (opposite of PC and propeller)
' The cycle redundancy check byte is sent in CRC7 or CRC16 format for all transactions
' except status tokens. CRC is of course a method of error detection, but not
' correction. If a CRC fails to match then the data must be sent or requested again.
' the algorithms for CRC7 and CRC16 are well documented and are basically division algorithms
' where the CRCs are the remainders. In our case, we are always going to send $95 which
' is the CRC for the first command that will be sent that matters, after that CRCs are
' ignored in SPI mode, later add code to compute the CRC for kicks if you like.
' Also, note the use of SD_RESPONSE_LIMIT, this allows the command to be sent over and over until there is a response.
' but, if the attempts exceed this constant, there is something wrong with the SD card.
'
' INPUTS: command8 - 8-bit command from SD card specification refer to document "SD Card Specification" from SD Association.
'         address32 - 32-bit address in normal little endian format, function converts to big endian before transport
'
' OUTPUTS: returns status of command to SD card, typically $FF means error, but for each command response can be different
' -----------------------------------------------------------------------------

  ' throw away read
  io.SPI_Send_Recv_Byte($FF)

  ' first send command
  io.SPI_Send_Recv_Byte(command8)

  ' now send 4-bytes of address in MS-Byte to LS-byte format
  io.SPI_Send_Recv_Byte( (address32 >> 24) & $FF)
  io.SPI_Send_Recv_Byte( (address32 >> 16) & $FF)
  io.SPI_Send_Recv_Byte( (address32 >> 8)  & $FF)
  io.SPI_Send_Recv_Byte( (address32 >> 0)  & $FF)

  ' finally the CRC
  io.SPI_Send_Recv_Byte( crc )

  ' now wait for response
  response := $FF ' default response is $FF which means error or time out

  repeat index from 0 to SD_RESPONSE_LIMIT
    ' read response by sending dummy data
    response := io.SPI_Send_Recv_Byte( $FF )

    ' test if we have a non-$FF response
    if (response <> $FF)
      repeat index2 from 0 to (response_length - 1)
        response_token[index2] := response
        ' last read that doesn't make it to token is intentional
        ' sends dummy data to allow SD card to finish operation,
        ' needed after all commands (and responses)
        response := io.SPI_Send_Recv_Byte( $FF )
      ' return first byte of results
      return ( response_token[0] )
  ' if we got here then something bad happened!
  return ( $FF )

' end SD_Send_Command(command8, address32)


PRI SD_Read_Data_Block(buffer, length) | response, index
  ' not to be confused with the 512 blocks of a card, this is a data transfer
  ' block and used in other areas as well. it is rather low level
  ' assumes a block with a data response has been sent, the caller
  ' has checked for errors, and the next transfer is the data block

  ' wait for the SD card to fetch the data from the flash storage
'  repeat index from 0 to SD_RETRIEVE_SECTOR_LIMIT
  repeat index from 0 to readTimeout
    response := io.SPI_Send_Recv_Byte($FF)
    ' test for proper response code $FE, break out of loop
    if (response < $10)        ' error response
      return (response)
    if (response == $FE)       ' data is ready
      quit

  ' return error if not responding
  if (response <> $FE)
    return ($FF) ' there was an error


  io.SPI_Read_Bytes(buffer, length)

  ' pull data off the wire to buffer
'  repeat index from 0 to length-1
'    byte[buffer][index] := io.SPI_Send_Recv_Byte($FF)

  ' next read in dummy data to finish
  io.SPI_Send_Recv_Byte($FF) ' ignore this data
  io.SPI_Send_Recv_Byte($FF) ' ignore CRC value
  io.SPI_Send_Recv_Byte($FF) ' finally 8 more clocks for the card to finish

  return ($00)


PRI SD_Write_Data_Block(buffer, length) | response
  ' not to be confused with the 512 blocks of a card, this is a data transfer
  ' block and used in other areas as well. it is rather low level
  ' assumes a block with a data followup has been sent, the caller
  ' has checked for errors, and the next transfer is the data block

  io.SPI_Send_Recv_Byte($FE)    ' Send start of data token.

  io.SPI_Write_Bytes(buffer, length)

'  repeat index from 0 to length-1
'    io.SPI_Send_Recv_Byte( byte[buffer][index] )

  ' send dummy crc
  ' FIXME: do real crc one day
  io.SPI_Send_Recv_Byte($FF) ' ignore this data
  io.SPI_Send_Recv_Byte($FF) ' ignore CRC value

  ' get response byte
  response := SD_Wait_Response
  response &= $1F         ' mask upper 5 bits per the spec
  if response <> $5
    return ($FF)
  ' all good
  return (0)

PRI SD_Wait_Response | index
  repeat index from 0 to 10
    result := io.SPI_Send_Recv_Byte($FF)
    if result <> $FF
      quit

PRI SD_Wait_Write_Busy | index, response
  ' assumes a command has been sent, probably a write, and the device is busy
  ' and unable to take commands.
  ' will poll it until it's no longer bust

  repeat index from 0 to writeTimeout
    response := io.SPI_Send_Recv_Byte($FF)
    ' test for proper response code $FE, break out of loop
    if (response <> $00)       ' non-zero means no longer busy
      return (0)
  ' fail
  return ($FF)

PRI Idle_Enter | sd_response, cmd_num_attempts
  ' Place SD card into idle state
  sd_response := $FF
  cmd_num_attempts := 0
  repeat while (sd_response <> $01)

    ' send SD command GO_IDLE_STATE = $40, which resets the SD card
    sd_response := SD_Send_Command( SD_GO_IDLE_STATE, 0, $95, 1 )

    ' check if we have tried this long enough
    if (++cmd_num_attempts > SD_RESPONSE_LIMIT)
      'Can't mount SD card, not responding to idle state command
      return false
  return true

' /////////////////////////////////////////////////////////////////////////////

PRI Idle_Exit_MMC | sd_response, cmd_num_attempts
  ' bring SD card out of idle state
  sd_response := $FF
 cmd_num_attempts := 0
  repeat while (sd_response <> $00)
    ' send SD command EXIT_IDLE_STATE = $41, which resets the SD card
    sd_response := SD_Send_Command( MMC_SEND_OP_COND, HCSBitMask, $95, 1 )
    ' check if we have tried this long enough
    if (++cmd_num_attempts > SD_RESPONSE_LIMIT)
        'SD card not responding to exit idle state command
      return false
  return true

PRI Idle_Exit_SD | sd_response, cmd_num_attempts
  ' bring SD card out of idle state
  sd_response := $FF
  cmd_num_attempts := 0
  repeat while (sd_response <> $00)
    ' send SD command EXIT_IDLE_STATE = $41, which resets the SD card
    sd_response := SD_Send_Command( SD_SEND_OP_COND, HCSBitMask, $95, 1 )
    ' check if we have tried this long enough
    if (++cmd_num_attempts > SD_RESPONSE_LIMIT)
        'SD card not responding to exit idle state command
      return false
  return true



PRI Interface_Condition
  ' sends a test pattern and voltage wishes to the SD card to test
  SD_Send_Command( SD_SEND_IF_COND, %1_1010_1010, $87, 5 )
  if (response_token[0] & $4)
      return 1                            ' command not found, SD 1.0 or non-SD
  if (response_token[3] & $1)             ' voltage ok, SD 2.0/3.0
    if (response_token[4] == $AA)         ' test pattern ok
      return 2                            ' ok signal
  return 0                               ' attempted but rejected condition

DAT
{{
                            TERMS OF USE: MIT License

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}
'//////////////////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////


